#include <sstream>
#include <GL/glew.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <SDL/SDL.h>
#include <SDL/SDL_ttf.h>
#include <SDL/SDL_image.h>
#include <math.h>
#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/time.h>

#include "glm-0.9.2.6/glm/glm.hpp"
#include "glm-0.9.2.6/glm/gtc/matrix_transform.hpp"
#include "glm-0.9.2.6/glm/gtc/type_ptr.hpp"

#include "src/vector.h"
#include "src/keyboard.h"
#include "src/timer.h"
#include "src/matrix.h"
#include "src/glhelper.h"
#include "src/obj.h"

#define MAX(x,y) ((x)>(y)?(x):(y))
#define MIN(x,y) ((x)<(y)?(x):(y))
#define CLAMP(x, y, z) ((x)<(y)?(y):((x)>(z)?(z):(x)))

enum geometry  { NONE, PLANE, SPHERE, TRIANGLE };
enum materials { DIFFUSE, SPECULAR, REFRACTIVE };

// saves the texture int the portable pixmap format
void savePPM(unsigned char* buffer, int width, int height, int bpp, const char* image_out) {
	std::fstream of(image_out, std::ios_base::out | std::ios_base::trunc);
	of << "P3" << std::endl;
	of << width << " " << height << std::endl;
	of << "255" << std::endl;
	for (int j = 0; j < height; j++) {
		for (int i = 0; i < width; i++) {
			of << (int)buffer[j * width * bpp / 8 + i * bpp / 8 + 0] << " "
			   << (int)buffer[j * width * bpp / 8 + i * bpp / 8 + 1] << " "
			   << (int)buffer[j * width * bpp / 8 + i * bpp / 8 + 2] << " ";
		}
		of << std::endl;
	}
}

// abstract base class: the plane, sphere, and triangle objects are derived from this
class cObject {
    public:
	int type;
	int material;
	vector3 color;
	vector3 emissive;

	cObject();
	cObject(vector3 color, vector3 emissive, int material);
	~cObject();
	virtual void applyCamera(cMatrix& camera) = 0;
	virtual void save() = 0;
};

cObject::cObject() : color(vector3(1.0,1.0,1.0)), emissive(vector3(0.0,0.0,0.0)), material(DIFFUSE) { }
cObject::cObject(vector3 color, vector3 emissive, int material) : color(color), emissive(emissive), material(material) { }
cObject::~cObject() { };//std::cout << "object destroyed" << std::endl; }


class cPlane : public cObject {
    public:
	vector3 normal, normal_;
	vector3 point, point_;
	
	cPlane();
	cPlane(vector3 normal, vector3 point, vector3 color, vector3 emissive, int material);
	~cPlane();
	void applyCamera(cMatrix& camera);
	void save();
};

cPlane::cPlane() : normal(vector3(0.0,0.0,1.0)), point(vector3()), normal_(vector3(0.0,0.0,1.0)), point_(vector3()), cObject(vector3(1.0,1.0,1.0), vector3(0.0,0.0,0.0), material) { this->type = PLANE; }
cPlane::cPlane(vector3 normal, vector3 point, vector3 color, vector3 emissive, int material) : normal(normal), point(point), normal_(normal), point_(point), cObject(color, emissive, material) { this->type = PLANE; }
cPlane::~cPlane() { std::cout << "plane destroyed" << std::endl; }
void cPlane::applyCamera(cMatrix& camera) {
	vector3 p = this->point_ + this->normal_;
	this->point = camera.mult(this->point_);
	p = camera.mult(p);
	this->normal = (p - this->point).unit();
}
void cPlane::save() {
	this->normal_ = this->normal;
	this->point_ = this->point;
}


class cSphere : public cObject {
    public:
	vector3 center, center_;
	double radius, radius_;

	cSphere();
	cSphere(vector3 center, double radius, vector3 color, vector3 emissive, int material);
	~cSphere();
	void applyCamera(cMatrix& camera);
	void save();
};

cSphere::cSphere() : center(vector3(0.0,0.0,0.0)), center_(vector3(0.0,0.0,0.0)), radius(1.0), radius_(1.0), cObject(vector3(1.0,1.0,1.0), vector3(0.0,0.0,0.0), material) { this->type = SPHERE; }
cSphere::cSphere(vector3 center, double radius, vector3 color, vector3 emissive, int material) : center(center), center_(center), radius(radius), radius_(radius), cObject(color, emissive, material) { this->type = SPHERE; }
cSphere::~cSphere() { std::cout << "sphere destroyed" << std::endl; }
void cSphere::applyCamera(cMatrix& camera) {
	this->center = camera.mult(this->center_);
}
void cSphere::save() {
	this->center_ = this->center;
	this->radius_ = this->radius;
}


class cTriangle : public cObject {
    public:
	vector3 p0, p1, p2, p0_, p1_, p2_;
	vector3 n0, n1, n2, n0_, n1_, n2_;
	vector3 n, n_;
	
	cTriangle();
	cTriangle(vector3 p0, vector3 p1, vector3 p2, vector3 n0, vector3 n1, vector3 n2, vector3 color, vector3 emissive, int material);
	~cTriangle();
	void applyCamera(cMatrix& camera);
	void save();
};

cTriangle::cTriangle() :
    p0(vector3(0.0,1.0,0.0)),
    p1(vector3(-1.0,0.0,0.0)),
    p2(vector3(1.0,0.0,0.0)),
    p0_(vector3(0.0,1.0,0.0)),
    p1_(vector3(-1.0,0.0,0.0)),
    p2_(vector3(1.0,0.0,0.0)),
    n0(vector3(0.0,0.0,1.0)),
    n1(vector3(0.0,0.0,1.0)),
    n2(vector3(0.0,0.0,1.0)),
    n0_(vector3(0.0,0.0,1.0)),
    n1_(vector3(0.0,0.0,1.0)),
    n2_(vector3(0.0,0.0,1.0)),
    cObject(vector3(1.0,1.0,1.0),vector3(0.0,0.0,0.0), material) {
	this->type = TRIANGLE;
	this->n = this->n_ = ((p1-p0).cross(p2-p0)).unit();
}
cTriangle::cTriangle(vector3 p0, vector3 p1, vector3 p2, vector3 n0, vector3 n1, vector3 n2, vector3 color, vector3 emissive, int material) :
    p0(p0),
    p1(p1),
    p2(p2),
    n0(n0),
    n1(n1),
    n2(n2),
    p0_(p0),
    p1_(p1),
    p2_(p2),
    n0_(n0),
    n1_(n1),
    n2_(n2),
    cObject(color, emissive, material) {
	this->type = TRIANGLE;
	this->n = this->n_ = ((p1-p0).cross(p2-p0)).unit();
}
cTriangle::~cTriangle() { std::cout << "plane destroyed" << std::endl; }
void cTriangle::applyCamera(cMatrix& camera) {
	vector3 _p0, _p1, _p2;
	vector3 _pn;

	_pn = this->p0_ + this->n_;
	_p0 = this->p0_ + this->n0_;
	_p1 = this->p1_ + this->n1_;
	_p2 = this->p2_ + this->n2_;

	this->p0 = camera.mult(this->p0_);
	this->p1 = camera.mult(this->p1_);
	this->p2 = camera.mult(this->p2_);

	_pn = camera.mult(_pn);
	_p0 = camera.mult(_p0);
	_p1 = camera.mult(_p1);
	_p2 = camera.mult(_p2);

	this->n  = (_pn - this->p0).unit();
	this->n0 = (_p0 - this->p0).unit();
	this->n1 = (_p1 - this->p1).unit();
	this->n2 = (_p2 - this->p2).unit();

}
void cTriangle::save() {
	this->p0_ = this->p0;
	this->p1_ = this->p1;
	this->p2_ = this->p2;
	this->n0_ = this->n0;
	this->n1_ = this->n1;
	this->n2_ = this->n2;
	this->n_  = this->n;
}


struct vertex_ {
	double x,  y,   z;
	double tx, ty, tz;
	double t[6];
};

class __vector {
  public:
	double x, y, z;

	__vector();
	__vector(double x, double y, double z);

	double dot(__vector w);
	double operator*(__vector w);
	__vector cross(__vector w);
	
	__vector add(__vector w);
	__vector operator+(__vector w);
	__vector sub(__vector w);
	__vector operator-(__vector w);
	__vector mult(double s);
	__vector operator*(double s);
	
	__vector h(__vector w);
	double length();
	__vector unit();
};

struct __ray {
	__vector origin;
	__vector direction;
};

struct __intersection {
	bool intersects;
	double t;
	
	__ray ray;
	__vector normal;
};

struct __object {
	int type;
	int material;
	__vector color;
	__vector emission;
	
	// for planes
	__vector normal;
	__vector point;
	
	// for spheres
	__vector center;
	double radius;
	
	// for triangles
	__vector p0, p1, p2;
	__vector n0, n1, n2;
	__vector n;
};

struct _bounding_box {
	unsigned short *depth,
		       *depth_device;
	double *minx, *miny, *minz, *maxx, *maxy, *maxz,
	       *minx_device, *miny_device, *minz_device, *maxx_device, *maxy_device, *maxz_device;
	short *child0, *child1,
	      *child0_device, *child1_device;
	short *parent,
	      *parent_device;

	short *id, *id_device;
	short *leaf_id, *leaf_id_device;

	unsigned short *n_objects,
		       *n_objects_device;
	unsigned short *objects,
		       *objects_device;
	unsigned short max_depth;
	unsigned short size, leaf_size;
};


bool initializePathTracer(_bounding_box& b, unsigned char*& current_buffer_host, unsigned char *& current_buffer_device, double*& current_doubles_device, __object* _objects, __object*& current_objects_device, int object_count, float*& rand_device, int max_bounces, int width, int height, double minx, double miny, double minz, double maxx, double maxy, double maxz);
bool setupPathTracer(double*& current_doubles_device, __object*& current_objects_device, __object* _objects, int object_count, int width, int height);
bool releasePathTracer(_bounding_box& b, unsigned char*& current_buffer_host, unsigned char *& current_buffer_device, double*& current_doubles_device, __object*& current_objects_device, float*& rand_device, int width, int height);
bool runPathTracer(_bounding_box& b, unsigned char*& current_buffer_host, unsigned char *& current_buffer_device, double*& current_doubles_device, float*& rand_device, unsigned long long samples, __object objects[], __object*& current_objects_device, int object_count, int max_bounces, int width, int height, int offsetx, int offsety, double focal_distance, double blur_radius);

int main(int argc, char* argv[]) {

	int WIDTH = 1280;
	int HEIGHT = 720;
	int BPP = 24;
	int MAX_BOUNCES = 5;
	
	double focal_distance = 1800.0; // updated below to center of min and max
	double blur_radius = 4;// 32.0;
	
	cKeyboard kb;
	cTimer t0; double elapsed0;
  
	timespec current;
	clock_gettime(CLOCK_REALTIME, &current);
	srand(current.tv_sec + current.tv_nsec / 1000000000.0);
	
	bool active = true;
	unsigned int samples = 1;

	SDL_Init(SDL_INIT_EVERYTHING);
	SDL_Surface *screen = mySDLInit(WIDTH, HEIGHT, BPP, false);
	SDL_Event event;

	std::vector<cObject*> objects;

	// two planes (sky and ground)
	objects.push_back(new cPlane(vector3( 0,-1, 0), vector3(   0, 8192,   0), vector3(  0,  0,  0), vector3(1,1,1), DIFFUSE)); // top
	objects.push_back(new cPlane(vector3( 0, 1, 0), vector3(   0,   20,   0), vector3(0.8,0.8,0.8), vector3(), DIFFUSE)); // bottom

	// load the obj model
	cObj obj2("monkey_subdivided.obj");
	for (int i = 0000; i < MIN(10000,obj2.faces.size()); i++) {
		objects.push_back(new cTriangle( vector3(obj2.vertices[obj2.faces[i].vertex[0]].v[0], obj2.vertices[obj2.faces[i].vertex[0]].v[1], obj2.vertices[obj2.faces[i].vertex[0]].v[2])*100+vector3(0,60,-70),
						 vector3(obj2.vertices[obj2.faces[i].vertex[1]].v[0], obj2.vertices[obj2.faces[i].vertex[1]].v[1], obj2.vertices[obj2.faces[i].vertex[1]].v[2])*100+vector3(0,60,-70),
						 vector3(obj2.vertices[obj2.faces[i].vertex[2]].v[0], obj2.vertices[obj2.faces[i].vertex[2]].v[1], obj2.vertices[obj2.faces[i].vertex[2]].v[2])*100+vector3(0,60,-70),
						 vector3(obj2.normals[obj2.faces[i].normal[0]].v[0], obj2.normals[obj2.faces[i].normal[0]].v[1], obj2.normals[obj2.faces[i].normal[0]].v[2]),
						 vector3(obj2.normals[obj2.faces[i].normal[1]].v[0], obj2.normals[obj2.faces[i].normal[1]].v[1], obj2.normals[obj2.faces[i].normal[1]].v[2]),
						 vector3(obj2.normals[obj2.faces[i].normal[2]].v[0], obj2.normals[obj2.faces[i].normal[2]].v[1], obj2.normals[obj2.faces[i].normal[2]].v[2]),
						 vector3(0.7,1.0,0.7), vector3(0,0.0,0), DIFFUSE) ); // top
	}

	// add some spheres resting on the ground
	for (int i = 0; i < 40; i++) {
		double sx = rand()/(double)RAND_MAX;
		double sy = rand()/(double)RAND_MAX;
		double sz = rand()/(double)RAND_MAX;
		double sr = rand()/(double)RAND_MAX;

		sx -= 0.5; sy -= 0.5; sz -= 0.5; sr += 0.5;
		sx *= 400.0*WIDTH/HEIGHT; sy *= 200; sz *= 400; sr *= 15;
		int type = rand()%2;

		sy = sr + 20; //100;
		double r = type == 0 ? rand()/(double)RAND_MAX : 0.9;
		double g = type == 0 ? rand()/(double)RAND_MAX : 0.9;
		double b = type == 0 ? rand()/(double)RAND_MAX : 0.9;

		objects.push_back(new cSphere(vector3(sx,sy,sz), sr, vector3(r, g, b), vector3(), type  )); // light source
	}

	// set up the camera and apply it to the primitives
	{
		vector3  origin(100,150,500);
		vector3 forward = (vector3(20,140,0)-origin).unit();
		vector3 up(0, 1,  0);
		vector3 right = (forward.cross(up)).unit();
			   up = (right.cross(forward)).unit();

		double cam[] = {  right.x,   right.y,   right.z, -(right*origin),
				     up.x,      up.y,      up.z, -(up*origin),
				forward.x, forward.y, forward.z, -(forward*origin),
					0,         0,         0,         1 };

		cMatrix camera(4, 4, cam);
		//camera.output();

		for (int i = 0; i < objects.size(); i++) {
			objects[i]->applyCamera(camera);
			objects[i]->save();
		}
	}

	double minx =  10000000, miny =  10000000, minz =  10000000;
	double maxx = -10000000, maxy = -10000000, maxz = -10000000;
	
	unsigned char *current_buffer_host;
	unsigned char *current_buffer_device;
	double *current_doubles_device;
	__object *current_objects_device;
	float *rand_device;
	_bounding_box b;

	// copy the transformed objects into an __object array
	__object _objects[objects.size()];
	{
		for (int i = 0; i < objects.size(); i++) {

			_objects[i].type = objects[i]->type;
			_objects[i].material = objects[i]->material;

			_objects[i].color.x = objects[i]->color.x;
			_objects[i].color.y = objects[i]->color.y;
			_objects[i].color.z = objects[i]->color.z;

			_objects[i].emission.x = objects[i]->emissive.x;
			_objects[i].emission.y = objects[i]->emissive.y;
			_objects[i].emission.z = objects[i]->emissive.z;

			if (objects[i]->type == PLANE) {
				_objects[i].normal.x = ((cPlane *)objects[i])->normal.x;
				_objects[i].normal.y = ((cPlane *)objects[i])->normal.y;
				_objects[i].normal.z = ((cPlane *)objects[i])->normal.z;
				_objects[i].point.x = ((cPlane *)objects[i])->point.x;
				_objects[i].point.y = ((cPlane *)objects[i])->point.y;
				_objects[i].point.z = ((cPlane *)objects[i])->point.z;
			} else if (objects[i]->type == SPHERE) {
				_objects[i].center.x = ((cSphere *)objects[i])->center.x;
				_objects[i].center.y = ((cSphere *)objects[i])->center.y;
				_objects[i].center.z = ((cSphere *)objects[i])->center.z;
				_objects[i].radius = ((cSphere *)objects[i])->radius;

				minx = MIN(minx,_objects[i].center.x-_objects[i].radius);
				miny = MIN(miny,_objects[i].center.y-_objects[i].radius);
				minz = MIN(minz,_objects[i].center.z-_objects[i].radius);
				maxx = MAX(maxx,_objects[i].center.x+_objects[i].radius);
				maxy = MAX(maxy,_objects[i].center.y+_objects[i].radius);
				maxz = MAX(maxz,_objects[i].center.z+_objects[i].radius);
			} else if (objects[i]->type == TRIANGLE) {
				_objects[i].p0.x = ((cTriangle *)objects[i])->p0.x;
				_objects[i].p0.y = ((cTriangle *)objects[i])->p0.y;
				_objects[i].p0.z = ((cTriangle *)objects[i])->p0.z;
				_objects[i].p1.x = ((cTriangle *)objects[i])->p1.x;
				_objects[i].p1.y = ((cTriangle *)objects[i])->p1.y;
				_objects[i].p1.z = ((cTriangle *)objects[i])->p1.z;
				_objects[i].p2.x = ((cTriangle *)objects[i])->p2.x;
				_objects[i].p2.y = ((cTriangle *)objects[i])->p2.y;
				_objects[i].p2.z = ((cTriangle *)objects[i])->p2.z;
				_objects[i].n0.x = ((cTriangle *)objects[i])->n0.x;
				_objects[i].n0.y = ((cTriangle *)objects[i])->n0.y;
				_objects[i].n0.z = ((cTriangle *)objects[i])->n0.z;
				_objects[i].n1.x = ((cTriangle *)objects[i])->n1.x;
				_objects[i].n1.y = ((cTriangle *)objects[i])->n1.y;
				_objects[i].n1.z = ((cTriangle *)objects[i])->n1.z;
				_objects[i].n2.x = ((cTriangle *)objects[i])->n2.x;
				_objects[i].n2.y = ((cTriangle *)objects[i])->n2.y;
				_objects[i].n2.z = ((cTriangle *)objects[i])->n2.z;
				_objects[i].n.x = ((cTriangle *)objects[i])->n.x;
				_objects[i].n.y = ((cTriangle *)objects[i])->n.y;
				_objects[i].n.z = ((cTriangle *)objects[i])->n.z;

				minx = MIN(minx,_objects[i].p0.x);
				miny = MIN(miny,_objects[i].p0.y);
				minz = MIN(minz,_objects[i].p0.z);
				maxx = MAX(maxx,_objects[i].p0.x);
				maxy = MAX(maxy,_objects[i].p0.y);
				maxz = MAX(maxz,_objects[i].p0.z);
				minx = MIN(minx,_objects[i].p1.x);
				miny = MIN(miny,_objects[i].p1.y);
				minz = MIN(minz,_objects[i].p1.z);
				maxx = MAX(maxx,_objects[i].p1.x);
				maxy = MAX(maxy,_objects[i].p1.y);
				maxz = MAX(maxz,_objects[i].p1.z);
				minx = MIN(minx,_objects[i].p2.x);
				miny = MIN(miny,_objects[i].p2.y);
				minz = MIN(minz,_objects[i].p2.z);
				maxx = MAX(maxx,_objects[i].p2.x);
				maxy = MAX(maxy,_objects[i].p2.y);
				maxz = MAX(maxz,_objects[i].p2.z);

				//std::cout << _objects[i].p0.x << " " << _objects[i].p0.y << " " << _objects[i].p0.z << std::endl;
				//std::cout << _objects[i].p1.x << " " << _objects[i].p1.y << " " << _objects[i].p1.z << std::endl;
				//std::cout << _objects[i].p2.x << " " << _objects[i].p2.y << " " << _objects[i].p2.z << std::endl;
			}
		}
	}

	std::cout << minx << " " << miny << " " << minz << " " << maxx << " " << maxy << " " << maxz << std::endl;
	
	focal_distance = sqrt(pow((minx+maxx)/2.0,2.0)+pow((miny+maxy)/2.0,2.0)+pow((minz+maxz)/2.0,2.0));
	focal_distance = 575.246;
	std::cout << "focal distance " << focal_distance << std::endl;

	// path tracer initialization
	initializePathTracer(b, current_buffer_host, current_buffer_device, current_doubles_device, _objects, current_objects_device, objects.size(), rand_device, MAX_BOUNCES, WIDTH, HEIGHT, minx, miny, minz, maxx, maxy, maxz);
	//setupPathTracer(current_doubles_device, current_objects_device, _objects, objects.size(), WIDTH, HEIGHT);

	// shaders
	GLuint glProgram, glShaderV, glShaderF;
	GLint vertex, texCoord, Projection, View, Model, textureSample;
	createProgram(glProgram, glShaderV, glShaderF, "src/vertex.sh", "src/fragment.sh");
	vertex        = glGetAttribLocation(glProgram, "vertex");
	texCoord      = glGetAttribLocation(glProgram, "texCoord");
	Projection    = glGetUniformLocation(glProgram, "Projection");
	View          = glGetUniformLocation(glProgram, "View");
	Model         = glGetUniformLocation(glProgram, "Model");
	textureSample = glGetUniformLocation(glProgram, "textureSample");

	// vertices
	const int vertex_count = 4;
	vertex_ vertices[vertex_count];
	vertices[0].x  = 0;     vertices[0].y  = 0;      vertices[0].z = 0;     vertices[0].tx = 0;     vertices[0].ty = 0;	
	vertices[1].x  = WIDTH; vertices[1].y  = 0;      vertices[1].z = 0;     vertices[1].tx = 1;     vertices[1].ty = 0;	
	vertices[2].x  = WIDTH; vertices[2].y  = HEIGHT; vertices[2].z = 0;     vertices[2].tx = 1;     vertices[2].ty = 1;
	vertices[3].x  = 0;     vertices[3].y  = HEIGHT; vertices[3].z = 0;     vertices[3].tx = 0;     vertices[3].ty = 1;
	GLuint vbo_vertices;
	glGenBuffers(1, &vbo_vertices);
	glBindBuffer(GL_ARRAY_BUFFER, vbo_vertices);
	glBufferData(GL_ARRAY_BUFFER, sizeof(vertex_) * vertex_count, vertices, GL_DYNAMIC_DRAW);

	// indices
	const int indices_count = 6;
	unsigned int indices[indices_count] = {0, 1, 2, 2, 3, 0};
	GLuint vbo_indices;
	glGenBuffers(1, &vbo_indices);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vbo_indices);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(unsigned int) * indices_count, indices, GL_STATIC_DRAW);

	// create a texture for rendering the result
	GLuint texture;
	setupTexture(texture);

	glm::mat4 projection = glm::ortho(0.0f, (float)WIDTH, (float)HEIGHT, 0.0f, -5.0f, 5.0f); 
	glm::mat4 view       = glm::mat4(1.0f);
	glm::mat4 model      = glm::mat4(1.0f);

	int offsetx = 0, offsety = 0;
	bool sleep = false;
	while (active) {

		while (SDL_PollEvent(&event)) {
			switch (event.type) {
			    case SDL_QUIT: active = false; break;
			}
		}

		elapsed0 = t0.elapsed(true);

		if (kb.getKeyState(KEY_SPACE))    { sleep = true; }
		if (kb.getKeyState(KEY_G))        { sleep = false; }
		
		if (sleep) {
			usleep(100000);
		} else {
			runPathTracer(b, current_buffer_host, current_buffer_device, current_doubles_device, rand_device, samples, _objects, current_objects_device, objects.size(), MAX_BOUNCES, WIDTH, HEIGHT, offsetx, offsety, focal_distance, blur_radius);
			std::cout << "samples: " << (int)samples << std::endl;

			glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

			glUseProgram(glProgram);

			// texture
			glUniform1i(textureSample, 0);
			glActiveTexture(GL_TEXTURE0);
			glBindTexture(GL_TEXTURE_2D, texture);
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, WIDTH, HEIGHT, 0, GL_RGB, GL_UNSIGNED_BYTE, current_buffer_host);

			// Project, View, and Model uniforms
			glUniformMatrix4fv(Projection, 1, GL_FALSE, glm::value_ptr(projection));
			glUniformMatrix4fv(View, 1, GL_FALSE, glm::value_ptr(view));
			model = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f,0.0f,-2.0f));
			glUniformMatrix4fv(Model, 1, GL_FALSE, glm::value_ptr(model));

			// vertex and texture attributes
			glBindBuffer(GL_ARRAY_BUFFER, vbo_vertices);
			glEnableVertexAttribArray(vertex);
			glVertexAttribPointer(vertex, 3, GL_DOUBLE, GL_FALSE, sizeof(vertex_), 0);
			glEnableVertexAttribArray(texture);
			glVertexAttribPointer(texCoord, 2, GL_DOUBLE, GL_FALSE, sizeof(vertex_), (char *)NULL + 24);

			// draw it
			glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vbo_indices);
			glDrawElements(GL_TRIANGLES, indices_count, GL_UNSIGNED_INT, 0);
			SDL_GL_SwapBuffers();
			
			offsetx += 64;
			if (offsetx > (WIDTH - 64)) {
				offsetx = 0;
				offsety += 64;
				if (offsety > (HEIGHT - 64)) {
					offsetx = offsety = 0;
					samples++;
				}
			}
		}

		if (samples % 10 == 0) {
			std::stringstream s;
			s << "screen" << (int)samples << ".ppm";
			savePPM(current_buffer_host, WIDTH, HEIGHT, BPP, s.str().c_str());//"screen.ppm");
		}

	}

	savePPM(current_buffer_host, WIDTH, HEIGHT, BPP, "screen.ppm");
	
	releasePathTracer(b, current_buffer_host, current_buffer_device, current_doubles_device, current_objects_device, rand_device, WIDTH, HEIGHT);

	// release the objects
	for (int i = 0; i < objects.size(); i++) delete objects[i];
	
	// release the texture, vertex buffer objects, and shader program.. also shut down SDL
	deleteTexture(texture);
	glDeleteBuffers(1, &vbo_indices);
	glDeleteBuffers(1, &vbo_vertices);
	releaseProgram(glProgram, glShaderV, glShaderF);
	SDL_Quit();

	return 0;
}